import csv
from saga.file_types.binary_file import parse_binary_file, write_binary_file
from saga.file_types.text_file import parse_text_file, write_text_file
from saga.file_types.csv_file import parse_csv_file, write_csv_file

def is_csv(file_path):
    csv_fileh = open(file_path, 'r')
    try:
        _ = csv.Sniffer().sniff(csv_fileh.read(1024))
        csv_fileh.seek(0)
        return True
    except csv.Error:
        return False


def parse_file(file_id, file_name, file_path):
    if file_path.endswith(".txt"):
        return parse_text_file(file_id, file_name, file_path)
    elif is_csv(file_path):
        return parse_csv_file(file_id, file_name, file_path)
    else:
        return parse_binary_file(file_id, file_name, file_path)


def write_file(file):
    if file.file_type == "csv":
        write_csv_file(file)
    elif file.file_type == "text":
        write_text_file(file)
    elif file.file_type == "binary":
        write_binary_file(file)
    else:
        print("Error: file type {} does not exist".format(file.file_type))


from saga.data_types.multi_dim_list.lcs import lcs

# diff3 merge
def diff3(A, O, B):
    matchings_A = lcs(A, O)
    matchings_B = lcs(B, O)

    # increment all the indexes by one
    matchings_A = [([a + 1], [o + 1], sim) for ([a], [o], sim) in matchings_A]
    matchings_B = [([b + 1], [o + 1], sim) for ([b], [o], sim) in matchings_B]

    chunks = get_chunks(A, O, B, matchings_A, matchings_B)

    string_output_chunks = chunks_to_output(A, O, B, matchings_A, matchings_B, chunks)

    return create_merged_file(string_output_chunks)

def get_chunks(A, O, B, matchings_A, matchings_B):
    chunks = []
    idx_A, idx_O, idx_B = 0, 0, 0
    max_idx_A, max_idx_O, max_idx_B = len(A), len(O), len(B)

    i = 1
    while not (i + idx_A > max_idx_A or i + idx_O > max_idx_O or i + idx_B > max_idx_B):
        if not is_matching(matchings_A, idx_A + i, idx_O + i) or not is_matching(matchings_B, idx_B + i, idx_O + i):
            # unstable 
            if i == 1:
                # find index end of unstable chunk
                for j in range(idx_O + 1, max_idx_O + 1):
                    last_indexes = is_stable_index(matchings_A, matchings_B, j)
                    if last_indexes:
                        # found last index of unstable chunk
                        (end_A, end_O, end_B) = last_indexes
                        unstable_chunk = ((idx_A + 1, end_A - 1), (idx_O + 1, end_O - 1), (idx_B + 1, end_B - 1))
                        chunks.append(unstable_chunk)
                        idx_A, idx_O, idx_B = end_A - 1, end_O - 1, end_B - 1
                        i = 1
                        break

            # stable
            else:
                stable_chunk = ((idx_A + 1, idx_A + i - 1), (idx_O + 1, idx_O + i - 1), (idx_B + 1, idx_B + i - 1))
                chunks.append(stable_chunk)
                idx_A, idx_O, idx_B = idx_A + i - 1, idx_O + i - 1, idx_B + i - 1
                i = 1
        else:
            i += 1
    else:
        last_chunk = ((idx_A + 1, max_idx_A), (idx_O + 1, max_idx_O), (idx_B + 1, max_idx_B))
        chunks.append(last_chunk)
        return chunks

def create_merged_file(chunks):
    merged_file = []
    conflicting_chunks = []
    for i, chunk in enumerate(chunks):
        (s_A, s_O, s_B) = chunk
        if not s_A == s_B:
            conflicting_chunks.append((i, chunk))
        else:
            merged_file.extend(s_A)

    if not any(conflicting_chunks):
        return merged_file
    return None

def chunks_to_output(A, O, B, matchings_A, matchings_B, chunks):
    calculated_ouput = []

    for chunk in chunks:

        ((s_A, e_A), (s_O, e_O), (s_B, e_B)) = chunk

        a_changed = chunk_changed(matchings_A, s_A, e_A, s_O, e_O)
        b_changed = chunk_changed(matchings_B, s_B, e_B, s_O, e_O)

        if a_changed and not b_changed:
            calculated_ouput.append(chunk_changed_in_A(chunk, A))
        elif b_changed and not a_changed:
            calculated_ouput.append(chunk_changed_in_B(chunk, B))
        else:
            calculated_ouput.append(chunk_conflicting_or_stable(chunk, A, B, O))

    return calculated_ouput

def chunk_changed_in_A(chunk, A):
    #output A[H], A[H], A[H]
    ((s_A, e_A), _, _) = chunk
    string = A[s_A - 1 : e_A]
    return (string, string, string)

def chunk_changed_in_B(chunk, B):
    #output B[H], B[H], B[H]
    (_, _, (s_B, e_B)) = chunk
    string = B[s_B  -1 : e_B]
    return (string, string, string)

def chunk_conflicting_or_stable(chunk, A, B, O):
    #output A[H], O[H], B[H]
    ((s_A, e_A), (s_O, e_O), (s_B, e_B)) = chunk
    return (A[s_A - 1: e_A], O[s_O - 1: e_O ], B[s_B - 1: e_B ])

def chunk_changed(matchings, s, e, s_O, e_O):

    for i in range (s, e + 1):
            if not matched(matchings, i):
                return True

    for i in range (s_O, e_O + 1):
            if not matched_O(matchings, i):
                return True

    return False

# returns true if a matching exists with this index of A or B
def matched(matchings, idx):
    for match in matchings:
        ([i], _, _) = match
        if (i == idx):
            return True
    return False

# returns true if a matching exists with this index of A or B
def matched_O(matchings, idx):
    for match in matchings:
        (_, [i], _) = match
        if (i == idx):
            return True
    return False

# i is matched to j
def is_matching(matchings, i, j):
    for match in matchings:
        (i_prime, j_prime, similarity) = match
        if i == i_prime[0] and j == j_prime[0]:
            return True
    return False

def is_stable_index(matchings_A, matchings_B, idx_O):
    for match_A in matchings_A:
        if match_A[1][0] == idx_O:
            for match_B in matchings_B:
                if match_B[1][0] == idx_O:
                    # return (idx_A, idx_O, idx_B)
                    return (match_A[0][0], match_A[1][0], match_B[0][0])
    return False 



from saga.base_file.File import File
from saga.data_types.multi_dim_list.MultiDimList import MultiDimList
import pandas as pd
from pandas import DataFrame
from pandas import ExcelWriter
from pandas import ExcelFile

def parse_excel_file(file_id, file_name, file_path):
    df = pd.read_excel(file_path, sheetname='Sheet1')
    l = MultiDimList(df.tolist(), 2)
    return File(file_id, "excel", file_name, l)

def write_excel_file(self, file):
    df = DataFrame.from_records(file.file_contents.multi_dim_list)
    writer = ExcelWriter(file.file_name)
    df.to_excel(writer,'Sheet1',index=False)
    writer.save()




import os
import pickle
import shutil
import hashlib
import copy
import filecmp
import glob
from os.path import join, isfile
from saga.Commit import Commit
from saga.file_types.file_utils import parse_file, write_file

class Repository(object):

    def __init__(self, directory):
        self.base_directory = directory
        self.saga_directory = self.base_directory + "/.saga/"
        self.commit_directory = self.base_directory + "/.saga/commits/"
        self.state_directory = self.base_directory + "/.saga/states/"
        self.index_directory = self.base_directory + "/.saga/index/"
        self.repo_pickle = self.base_directory + "/.saga/repository"
        self.head = "master"
        self.branches = {"master" : None} # map from branch name -> commit hash
        self.index = {"master" : set()} # map from branch name -> set of files we are tracking
        self.file_ids = {"master": dict()} # map from branch -> file_name -> permanent_file_id
        self.commits = {"master": []} # this is an ugly hack for now, but we keep track of commit hashes on any branch

    @staticmethod
    def init(directory):
        # we make a new .saga repository, if possible
        if os.path.exists(directory + "/.saga"):
            raise ValueError("There is already a saga project at {}".format(directory))
        os.mkdir(directory + "/.saga")
        os.mkdir(directory + "/.saga/commits")
        os.mkdir(directory + "/.saga/states")
        os.mkdir(directory + "/.saga/index")
        repo = Repository(directory)
        repo.commit("Create repository") # we add this empty commit so we don't need special cases
        return repo

    @staticmethod
    def read(directory):
        # read from the saga repo pickle
        repo_pickle = directory + "/.saga/repository"
        f = open(repo_pickle, "rb")
        p = f.read()
        f.close()
        return pickle.loads(p)

    def write(self):
        # read from the saga repo pickle
        f = open(self.repo_pickle, "wb+")
        p = pickle.dumps(self)
        f.write(p)
        f.close()

    def get_commit(self, commit_hash):
        f = open(self.commit_directory + commit_hash, "rb")
        commit_bytes = f.read()
        f.close()
        return Commit.from_bytes(commit_bytes)

    def _add_commit_to_db(self, commit):
        # if the commit exists, we don't need to error
        commit_hash = commit.hash()
        commit_bytes = commit.to_bytes()
        f = open(self.commit_directory + commit_hash, "wb+")
        f.write(commit_bytes)
        f.close()
        return commit_hash

    def index_hash(self):
        file_hashes = []
        # we sort to have the same files every time
        for file_name in sorted(self._relative_paths_in_dir(self.index_directory)):
            index_file_name = join(self.index_directory, file_name)
            if os.path.isfile(index_file_name):
                f = open(index_file_name, "rb")
                file_bytes = f.read()
                m = hashlib.sha256()
                m.update(file_bytes) # we encode the file contents
                m.update(file_name.encode("utf-8")) # and we encode the file path
                file_hashes.append(m.hexdigest())
        m = hashlib.sha256()
        #TODO: this should be a merkle tree eventually 
        # (or maybe an merkel-mountain-range), to reap the benefits
        m.update(",".join(file_hashes).encode('utf-8')) 
        return m.hexdigest()

    def get_new_file_id(self):
        # generates a random ID for a file
        import uuid
        return uuid.uuid4().hex

    def add(self, path):
        if not os.path.exists(path):
            print("Error: path {} does not exist".format(path))

        self._try_create_file_ids([path])

        if os.path.isfile(path):
            self._copy_file_to_dir(path, self.index_directory)
        else:
            assert os.path.isdir(path)
            # make ids for all files in directory
            files_in_path = self._relative_paths_in_dir(path)
            files_in_path = [join(path, f) for f in files_in_path]
            self._try_create_file_ids(files_in_path)
            # copy directory to index
            self._copy_dir_to_dir(path, join(self.index_directory, path))

    def _try_create_file_ids(self, file_paths):
        for path in file_paths:
            if path not in self.file_ids[self.head]:
                self.file_ids[self.head][path] = self.get_new_file_id()

    def commit(self, commit_message):
        state_hash = self.index_hash()
        parent_commit_hash = self.branches[self.head]
        commit = Commit(state_hash, [parent_commit_hash], commit_message)
        commit_hash = self._add_commit_to_db(commit)
        print("[{}] {}".format(commit_hash[0:12], commit_message))
        self.commits[self.head].append(commit_hash)
        self._backup_index_to_db()
        self.branches[self.head] = commit_hash

    def create_branch(self, branch_name):
        if branch_name in self.branches:
            print("Error: branch {} already exists".format(branch_name))
        self.branches[branch_name] = self.branches[self.head]
        self.index[branch_name] = copy.deepcopy(self.index[self.head])
        self.file_ids[branch_name] = copy.deepcopy(self.file_ids[self.head])
        self.commits[branch_name] = copy.deepcopy(self.commits[self.head])

    def switch_to_branch(self, branch_name):
        if branch_name not in self.branches:
            print("Error: cannot switch to branch {} as it does not exist".format(branch_name))
            return

        inserted, changed, removed = self.changed_files(self.curr_state_dir(), self.index_directory)
        if any(inserted) or any(changed) or any(removed):
            print("Error: please commit changes before switching branches")
            return

        self.head = branch_name
        commit = self.get_commit(self.branches[self.head])
        self._restore_state(commit.state_hash)        

    def _relative_paths_in_dir(self, directory, ignore=None):    
        if ignore is None:
            ignore = []

        # we make sure that the path does not start with a slash
        if directory[-1] == "/":
            paths = [path[len(directory):] for path in glob.glob(directory + '/**', recursive=True) if path[len(directory):] != ""]
        else:
            paths = [path[len(directory) + 1:] for path in glob.glob(directory + '/**', recursive=True) if path[len(directory) + 1:] != ""]
        def include_path(path):
            for ignore_path in ignore:
                if path.startswith(ignore_path):
                    return False
            if path == directory:
                return False
            return True
        return {path for path in paths if include_path(path)}

    def state_hash(self):
        return self.get_commit(self.branches[self.head]).state_hash

    def curr_state_dir(self):
        return join(self.state_directory, self.get_commit(self.branches[self.head]).state_hash)

    def changed_files(self, dir1, dir2):
        previous_state = self._relative_paths_in_dir(dir1)
        current_state = self._relative_paths_in_dir(dir2)

        removed_paths, changed_paths, inserted_paths  = set(), set(), set()
        for path in previous_state:
            if path not in current_state:
                removed_paths.add(path)
            else:
                path1 = join(dir1, path)
                path2 = join(dir2, path)
                if isfile(path1) and isfile(path2) and not filecmp.cmp(path1, path2):
                    changed_paths.add(path)
                elif (isfile(path1) and not isfile(path2)) or (not isfile(path1) and isfile(path2)):
                    changed_paths.add(path)
        for path in current_state:
            if path not in previous_state:
                inserted_paths.add(path)

        return removed_paths, changed_paths, inserted_paths

    def log(self):

        for commit_hash in reversed(self.commits[self.head]):
            commit = self.get_commit(commit_hash)
            print("commit: {}".format(commit_hash))
            print("\t{}\n".format(commit.commit_message))
        
    def status(self):
        print("On branch {}".format(self.head))

        rem_state_to_index, cha_state_to_index, ins_state_to_index = self.changed_files(self.curr_state_dir(), self.index_directory)
        rem_index_to_curr, cha_index_to_curr, ins_index_to_curr = self.changed_files(self.index_directory, self.base_directory)

        print("Changes staged for commit:")
        for path in rem_state_to_index:
            print("\tremoved: {}".format(path))
        for path in ins_state_to_index:
            print("\tinserted: {}".format(path))
        for path in cha_state_to_index:
            if path not in cha_index_to_curr and path not in rem_index_to_curr:
                print("\tmodified: {}".format(path))

        print("Change not staged for commit:")
        for path in rem_index_to_curr:
            print("\tremoved: {}".format(path))
        for path in cha_index_to_curr:
            print("\tmodified: {}".format(path))

        print("Untracked files:")
        for path in ins_index_to_curr:
            print("\t{}".format(path))


    def diff(self):
        _, changed_paths, _ = self.changed_files(self.index_directory, self.base_directory)

        print("Git diff:")
        operations = []
        for path in changed_paths:
            if os.path.isfile(path):
                print("\tFile:", path)
                # we should read in the current file and the old file in this case
                file_id = self.file_ids[self.head][path]
                old_file = parse_file(file_id, path, join(self.index_directory, path))
                new_file = parse_file(file_id, path, join(self.base_directory, path))
                ops = old_file.get_operations(new_file)
                for op in ops:
                    print("\t\t:", op)
                operations.extend(ops)
            else:
                print("\tDirectory:", path, "changed")

        return operations

    def branch(self):
        for branch in self.branches:
            if branch == self.head:
                print("   * {}".format(branch))
            else:
                print("     {}".format(branch))

    def merge(self, other_branch):
        if other_branch not in self.branches:
            print("Error: cannot merge branch {} as it does not exist".format(other_branch))

        lca, need_merge = self.least_common_ancestor(self.head, other_branch)

        if not need_merge:
            if len(self.commits[other_branch]) > len(self.commits[self.head]):
                self.branches[self.head] = self.branches[other_branch]
                self.commits[self.head] = copy.deepcopy(self.commits[other_branch])
                commit = self.get_commit(self.branches[self.head])
                self._restore_state(commit.state_hash)
                print("Fast-forward branch {} to branch {}".format(self.head, other_branch))
            else:
                print("Branch {} is a subbranch of {}".format(other_branch, self.head))
        else:

            state_dir_old = join(self.state_directory, self.get_commit(lca).state_hash)
            state_dir_newA = join(self.state_directory, self.state_hash())
            state_dir_newB = join(self.state_directory, self.get_commit(self.branches[other_branch]).state_hash)

            removed_paths_A, changed_paths_A, inserted_paths_A = self.changed_files(state_dir_old, state_dir_newA)
            removed_paths_B, changed_paths_B, inserted_paths_B = self.changed_files(state_dir_old, state_dir_newB)

            possibly_mergable = set() 
            print(inserted_paths_A)
            print(inserted_paths_B)
            for path in removed_paths_A.intersection(changed_paths_B):
                print("Confict as {} was removed in {} and modified in {}".format(path, self.head, other_branch))
            for path in removed_paths_B.intersection(changed_paths_A):
                print("Confict as {} was removed in {} and modified in {}".format(path, self.head, other_branch))
            for path in changed_paths_A.intersection(changed_paths_B):
                possibly_mergable.add(path)
                #print("Conflict as {} was modified in both {} and {}".format(path, self.head, other_branch))
            for path in inserted_paths_A.intersection(inserted_paths_B):
                print("Conflict as {} was inserted in both {} and {}".format(path, self.head, other_branch))

            files_to_write = [] # we only write the files if there are no merge conflicts in any of them
            for path in possibly_mergable:
                if path not in self.file_ids[self.head]: # we aren't tracking it
                    continue

                file_id = self.file_ids[self.head][path]
                file_o = parse_file(file_id, path, join(state_dir_old, path))
                file_a = parse_file(file_id, path, join(state_dir_newA, path))
                file_b = parse_file(file_id, path, join(state_dir_newB, path))

                merge_file = file_o.merge(file_a, file_b)
                if merge_file is None:
                    print("Merge conflict for file {}".format(file_o.file_name))
                    return
                else:
                    print("Successfully merged file {}".format(file_o.file_name))
                    files_to_write.append(merge_file)
            for f in files_to_write:
                write_file(f)
            self.commit("Merged {} into branch {}".format(other_branch, self.head))

    def least_common_ancestor(self, branch_1, branch_2):
        branch_1_commits = self.commits[branch_1]
        branch_2_commits = self.commits[branch_2]

        for i in reversed(range(min(len(branch_1_commits), len(branch_2_commits)))):
            if branch_1_commits[i] == branch_2_commits[i]:
                if i == min(len(branch_1_commits), len(branch_2_commits)) - 1:
                    return branch_1_commits[i], False
                else:
                    return branch_1_commits[i], True 
        
        raise Exception("No common ancestor")

    def _backup_index_to_db(self):
        # copies all files that are being tracked to a state hash directory
        dst = join(self.state_directory, self.index_hash())
        if not os.path.isdir(dst):
            os.mkdir(dst)
            # TODO: make this work when the cwd isn't the base repository
            self._copy_dir_to_dir(".saga/index/", dst)
        

    def _restore_state(self, state_hash):
        print("HERE")
        self._copy_dir_to_dir(join(".saga/states", state_hash), ".")


    def _copy_dir_to_dir(self, src, dst):
        """
        src is a relative path. It must be a folder that exists.
        dst is an absolute path. It will be created if it does not exist.

        If a path = src/file_name, then this will be copied to dst/file_name
        """
        if src[-1] == "/":
            src = src[:-1]

        if not os.path.exists(dst):
                os.makedirs(dst)

        # otherwise, we recursively expore the directory and copy it over
        for root, dirs, files in os.walk(src):
            relative_root = root[len(src) + 1:]

            # first we copy the directories
            for directory in dirs:
                # if the directory doesn't exist, we make it
                path = join(dst, relative_root, directory)
                if not os.path.isdir(path):
                    os.mkdir(path)

            # and then the files
            for file_name in files:
                path = join(dst, relative_root, file_name)
                shutil.copyfile(join(root, file_name), path)

    def _copy_file_to_dir(self, src, dst):
        """
        src is a relative path of a file. 
        dst is an absolute path of a folder. 

        src will eixst at join(dst, src)
        """

        dirname = join(dst, os.path.dirname(src))
        print(dirname)
        if not os.path.isdir(dirname):
            os.makedirs(dirname)

        shutil.copyfile(src, join(dst, src))


from copy import deepcopy
from saga.Operation import Operation

class OP_File_ChangeName(Operation):

    def __init__(self, file_id, old_name, new_name):
        Operation.__init__(self, "OP_File_ChangeName", file_id)
        self.old_name = old_name
        self.new_name = new_name

    def apply_operation_to_file(self, file):
        # TODO: figure out how we might extend this notion of invalidity to paths in list:
        # because we might insert a bunch of lines at the start of a file,
        # the old path might point to different values (and it would still be valid)
        if file.file_name != self.old_name:
            raise Exception("Invalid operation")

        file.file_name = self.new_name
        return file

    def inverse(self):
        return OP_File_ChangeName(self.file_id, self.new_name, self.old_name)

    def __str__(self):
        return "OP_File_ChangeName: changed file {} to {}".format(self.old_name, self.new_name)


from saga.data_types.multi_dim_list.lcs import changed_paths, inserted_paths, removed_paths, lcs_multi_dimension, get_matching_path
from saga.data_types.multi_dim_list.OP_MDL_Change import OP_MDL_Change
from saga.data_types.multi_dim_list.OP_MDL_Insert import OP_MDL_Insert
from saga.data_types.multi_dim_list.OP_MDL_Remove import OP_MDL_Remove
from saga.data_types.multi_dim_list.mdl_merge_utils import diff3

def value_at_path(arr, path):
    if len(path) == 0:
        return arr
    if len(path) == 1:
        return [arr[path[0]]]
    elif path[0] == "_":
        values = []
        for sublist in arr:
            values.extend(value_at_path(sublist, path[1:]))
        return values
    else:
        return value_at_path(arr[path[0]], path[1:])
    

class MultiDimList(object):

    def __init__(self, multi_dim_list, dimension):
        self.multi_dim_list = multi_dim_list
        self.dimension = dimension

    def remove_path(self, path):
        self._remove_path_rec(self.multi_dim_list, path)

    def _remove_path_rec(self, arr, path):
        if len(path) == 0:
            arr.clear()
            return
        if len(path) == 1:
            del arr[path[0]]
        elif path[0] == "_":
            for sublist in arr:
                self._remove_path_rec(sublist, path[1:])
        else:
            self._remove_path_rec(arr[path[0]], path[1:])
        
    def insert_path(self, path, value):
        self.multi_dim_list = self._insert_path_rec(self.multi_dim_list, path, value) 

    def _insert_path_rec(self, arr, path, value):
        if len(path) == 0:
            return value

        if len(path) == 1:
            arr.insert(path[0], value[0])
        elif path[0] == "_":
            step = int(len(value) / len(arr)) 
            for idx, sublist in enumerate(arr):
                arr[idx] = self._insert_path_rec(sublist, path[1:], value[idx * step: idx * step + step])
        else:
            arr[path[0]] = self._insert_path_rec(arr[path[0]], path[1:], value)
        
        return arr

    def change_value(self, path, value):
        self._change_value_rec(self.multi_dim_list, path, value)

    def _change_value_rec(self, arr, path, value):
        if len(path) == 1:
            arr[path[0]] = value
        else:
            self._change_value_rec(arr[path[0]], path[1:], value)

    def get_operations(self, multi_dim_obj):
        A = self.multi_dim_list
        B = multi_dim_obj.multi_dim_list

        dimension_matches = lcs_multi_dimension(A, B, self.dimension)

        operations = []

        removed_rows, removed_cols = removed_paths(A, B, dimension_matches)
        for path in removed_rows + removed_cols:
            operations.append(OP_MDL_Remove("id", path, value_at_path(A, path)))

        inserted_rows, inserted_cols = inserted_paths(A, B, dimension_matches)
        for path in inserted_rows + inserted_cols:
            operations.append(OP_MDL_Insert("id", path, value_at_path(B, path)))

        changed = changed_paths(A, B, dimension_matches)
        for path in changed:
            a_path, _ = get_matching_path(dimension_matches, False, path)
            operations.append(OP_MDL_Change("id", path, value_at_path(A, a_path), value_at_path(B, path)))
        
        return operations

    def merge(self, a_mdl, b_mdl):
        merge_res = diff3(a_mdl.multi_dim_list, self.multi_dim_list, b_mdl.multi_dim_list)
        if merge_res is None:
            return None

        return MultiDimList(merge_res, self.dimension)

    


from saga.Operation import Operation

class OP_MDL_Remove(Operation):

    def __init__(self, file_id, path, value):
        Operation.__init__(self, "OP_MDL_Remove", file_id)
        self.path = path
        self.value = value

    def apply_operation_to_file(self, file):
        if self.file_id != file.file_id:
            raise ValueError("Wrong file provided")

        file.file_contents.remove_path(self.path)
        return file

    def inverse(self):
        from saga.data_types.multi_dim_list.OP_MDL_Insert import OP_MDL_Insert 
        return OP_MDL_Insert(self.file_id, self.path, self.value)

    def __str__(self):
        return "OP_MDL_Remove: at path {}, removed {}".format(self.path, self.value)


"""
A set of functions for computing longest-common subsequences between lists.
"""
from copy import deepcopy


# Standard LCS
def lcs(A, B):
    def equal(a, b):
        if a == b:
            return 1
        return 0

    return lcs_similarity(A, B, equal)

# LCS with a similarity function. 
def lcs_similarity(A, B, similarity_function):
    m = len(A) 
    n = len(B) 

    # L[i][j] is the highest similar metric of the close longest common subsequece up to indexes i, j
    L = [[None]*(n + 1) for i in range(m + 1)] 
    for i in range(m + 1): 
        for j in range(n + 1): 
            if i == 0 or j == 0 : 
                L[i][j] = 0
            else:
                L[i][j] = max(
                    similarity_function(A[i - 1], B[j - 1]) + L[i - 1][j - 1],
                    L[i - 1][j],
                    L[i][j - 1]
                )

    # Create an array to store indexes of close common subsequence
    matches = []

    # Start from the right-most-bottom-most corner and 
    # one by one store characters in lcs[] 
    i = m 
    j = n 
    while i > 0 and j > 0:

        # If current character in X[] and Y are same, then 
        # current character is part of LCS 
        if similarity_function(A[i - 1], B[j - 1]) + L[i - 1][j - 1] > max(L[i - 1][j], L[i][j - 1]): 
            matches.append(([i - 1], [j - 1], similarity_function(A[i - 1], B[j - 1])))
            i-=1
            j-=1

        # If not same, then find the larger of two and 
        # go in the direction of larger value 
        elif L[i-1][j] > L[i][j-1]:
            i-=1
        else: 
            j-=1

    return matches


def arr_equals(A, B):
    if len(A) != len(B):
        return False
    if type(A) in (int, float, bool, str) and type(B) in (int, float, bool, str):
        return A == B
    for a, b in zip(A, B):
        if not arr_equals(a, b):
            return False
    return True

def list_from_path(matrix, path):
    curr = matrix
    for step in path:
        curr = curr[step]
    return curr

def list_similiarity_function(A, B):
    if type(A) in (int, float, bool) and type(B) in (int, float, bool):
        if A == B:
            return 1
        else:
            return 0
    elif len(A) == 0 and len(B) == 0:
        return 1
    elif type(A) == str and type(B) == str:
        indexes = lcs(A, B)
        if any(indexes):
            return len(indexes) / max(len(A), len(B))
        return 0

    indexes = lcs_similarity(A, B, list_similiarity_function)
    if any(indexes):
        total = sum(sim for _, _, sim in indexes)
        return total / max(len(A), len(B))
    return 0

# returns a mapping from "dimension down" to "matches at that level"
def lcs_multi_dimension(A, B, dimension):    
    dimension_matches = {i : [] for i in range(1, dimension + 1)}

    dimension_matches[1] = lcs_similarity(A, B, list_similiarity_function)

    for dimension_down in range(1, dimension):
        for path_a, path_b, _ in dimension_matches[dimension_down]:
            list_a = list_from_path(A, path_a)
            list_b = list_from_path(B, path_b)

            matches = lcs_similarity(list_a, list_b, list_similiarity_function)
            for idx_a, idx_b, sim in matches:
                dimension_matches[dimension_down + 1].append((path_a + idx_a, path_b + idx_b, sim))

    return dimension_matches

def get_matching_path(dim_matches, path_is_A, path):
    for path_a, path_b, sim in dim_matches[len(path)]:
        if path_is_A:
            if path_a == path:
                return path_b, sim
        else:
            if path_b == path:
                return path_a, sim
    return None, None


def inserted_paths(A, B, dim_matches):
    return inserted_paths_rec(A, B, dim_matches, False, [])

"""
Returns a tuple of (inserted rows, inserted columns):
inserted rows is a list of raw paths (also just lists)
the last index in an inserted column is not the row that was inserted but the column
"""
def inserted_paths_rec(A, B, dim_matches, first_list, base_path):
    if len(base_path) == max(dim_matches):
        return ([], [])

    # if this current thing was inserted
    lower_level_inserted_rows = []
    lower_level_inserted_cols = []
    add_curr_list = True
    indices = reversed(range(len(B))) if first_list else range(len(B)) # we go backwards if we are removing, so we don't have to shift
    for idx in indices:
        # if a path is matched, then things may have been inserted below it
        matching_path, _ = get_matching_path(dim_matches, first_list, base_path + [idx])
        if matching_path is not None:
            inserted_rows, inserted_cols = inserted_paths_rec(A[matching_path[-1]], B[idx], dim_matches, False, base_path + [idx])
            lower_level_inserted_rows.extend(inserted_rows)
            lower_level_inserted_cols.extend(inserted_cols)
            add_curr_list = False
        else:
            lower_level_inserted_rows.append(base_path + [idx])

    if add_curr_list:
        # we don't need to add anything below it, we can just add it as a whole
        return ([base_path], [])
    else:
        # now we try and see if these rows have a new column being created (happens at the bottom)
        two_deep_insertions = [x for x in lower_level_inserted_rows if len(x) == len(base_path) + 2]
        num_inserted_in_column = {path[-1] : 0 for path in two_deep_insertions} 
        for path in two_deep_insertions:
            num_inserted_in_column[path[-1]] += 1

        # if we also inserted a whole row, then we inserted all elements of that column a well
        one_deep_insertions = [x for x in lower_level_inserted_rows if len(x) == len(base_path) + 1]
        for path in one_deep_insertions:
            if len(path) < max(dim_matches):
                for col in range(len(B[path[-1]])):
                    if col not in num_inserted_in_column:
                        num_inserted_in_column[col] = 0
                    num_inserted_in_column[col] += 1
        
        for column in num_inserted_in_column:
            # find the number of rows that contain this column, as the lists are not rectangular
            column_length = sum(1 for row in B if column < len(row))

            if num_inserted_in_column[column] == column_length:
                lower_level_inserted_cols.append(base_path + ["_", column])
                # clear insertions from rows
                for path in two_deep_insertions:
                    if path[-1] == column:
                        lower_level_inserted_rows.remove(path)

        # and then we see if there are any columns we can unify
        to_delete = []
        for path in lower_level_inserted_cols:
            if len(path) >= len(base_path) + 2 and path[len(base_path) + 1] == "_":
                to_delete.append(path)

        if len(to_delete) == len(B):
            # we can unify 
            new_path = deepcopy(to_delete[0])
            new_path[len(base_path)] = "_"
            lower_level_inserted_cols.append(new_path)

            for path in to_delete:
                lower_level_inserted_cols.remove(path)

        return (lower_level_inserted_rows, lower_level_inserted_cols)


def removed_paths(A, B, dim_matches):
    # A deleted path is also a path that was added from the new list to the old list
    return inserted_paths_rec(B, A, dim_matches, True, [])

def changed_paths(A, B, dim_matches):
    changed = []
    for path in lowest_dim_paths(B, max(dim_matches), []):
        matching_path, sim = get_matching_path(dim_matches, False, path)
        if sim is not None and sim < 1:
            changed.append(path)

    return changed


def lowest_dim_paths(B, dim, base_path):
    if dim == 1:
        return [base_path + [idx] for idx in range(len(B))]
    
    paths = []
    for idx in range(len(B)):
        paths.extend(lowest_dim_paths(B[idx], dim - 1, base_path + [idx]))
    return paths





    




from saga.Operation import Operation

class OP_MDL_Change(Operation):

    def __init__(self, file_id, path, old_value, new_value):
        Operation.__init__(self, "OP_MDL_Change", file_id)
        self.path = path
        self.old_value = old_value
        self.new_value = new_value

    def apply_operation_to_file(self, file):
        if self.file_id != file.file_id:
            raise ValueError("Wrong file provided")

        file.file_contents.change_value(self.path, self.new_value)
        return file

    def inverse(self):
        return OP_MDL_Change(self.file_id, self.path, self.new_value, self.old_value)

    def __str__(self):
        return "OP_MDL_Change: at path {}, changed {} to {}".format(self.path, self.old_value, self.new_value)

    


from copy import deepcopy

class Operation():

    def __init__(self, op_name, file_id):
        self.op_name = op_name
        self.file_id = file_id

    def apply_operation(self, state):
        try:
            file = state[self.file_id]
            state[self.file_id] = self.apply_operation_to_file(file)
            return deepcopy(state)
        except:
            return None

    def apply_operation_to_file(self, file):
        """
        Returns a modified version of the file, with the ooeration applied
        """
        raise NotImplementedError("Must be implemented by specific operation")


    def inverse(self):
        """
        Returns an operation that when applied after this operation will result in the
        initial state
        """
        raise NotImplementedError("Must be implemented by specific operation")

    # TODO: come up some some way of storing these data blobs, maybe lik git does. 

    def to_string(self):
        """
        Returns the operation written to a string
        """
        raise NotImplementedError("Must be implemented by specific operation")


    @staticmethod
    def from_string(operation_string):
        """
        Returns an instance of the operation parsed from a valid operation string
        """
        raise NotImplementedError("Must be implemented by specific operation")



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="saga-vcs",
    version="0.0.8",
    author="saga",
    author_email="narush@wharton.upenn.edu.com",
    description="saga is a version control CLI that handles many file formats",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/naterush/saga",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
    ],
     entry_points={
        'console_scripts': [
            'saga = saga.main:main'
        ]
    },
    python_requires='>=3.6',
)


from saga.base_file.OP_File_ChangeName import OP_File_ChangeName

class File(object):

    def __init__(self, file_id, file_type, file_name, file_contents):
        """
        Create a new representation of a file:
            - file_id, file_type are static and can never change
            - file_name, file_contents can
        """
        self.file_id = file_id
        self.file_type = file_type
        self.file_name = file_name
        self.file_contents = file_contents


    def get_operations(self, other_file):
        """
        Get the operations between this file and other_file, which is thought of as the new
        version of the file. 
        """
        #if other_file is None:
        #    return Insert -- how to do delete?

        if self.file_id != other_file.file_id:
            raise ValueError("Can only get operations between different versions of the same file")

        operations = []
        if self.file_name != other_file.file_name:
            operations.append(OP_File_ChangeName(self.file_id, self.file_name, other_file.file_name))

        # now we delegate the operations to the file_contents
        operations.extend(self.file_contents.get_operations(other_file.file_contents))
        return operations

    # this should be called on the origin file
    def merge(self, file_a, file_b):
        merge_res = self.file_contents.merge(file_a.file_contents, file_b.file_contents)
        if merge_res is None:
            return None
        return File(self.file_id, self.file_type, self.file_name, merge_res)


from copy import deepcopy
from saga.Operation import Operation
from saga.base_file.File import File

class OP_File_Insert(Operation):

    def __init__(self, file_id, file_type, file_name, file_contents):
        Operation.__init__(self, "OP_File_Insert", file_id)
        self.file_type = file_type
        self.file_name = file_name
        self.file_contents = file_contents

    def apply_operation_to_file(self, file):
        if file is not None:
            raise Exception("Invalid operation")

        return File(self.file_id, self.file_type, self.file_name, self.file_contents)

    def inverse(self):
        from saga.base_file.OP_File_Remove import OP_File_Remove
        return OP_File_Remove(self.file_id, self.file_type, self.file_name, self.file_contents)

    def __str__(self):
        return "OP_File_Insert: inserted file {} of type {}".format(self.file_name, self.file_type)